            include    aoaExcepts.inc

; genericSEHfltr-
;
; This procedure must be a LEAF function!
; This function must not modify any register values other
; than EAX (the return value).
;
; Passed the following arguments:
;
;    RCX:   RIP value of interrupted instruction.
;    EDX:   Exception code.
;    R8:    Pointer to scopeTable record that
;           this function is to test.
;
; Return values (in RAX):
;
;    ExceptionContinueExecution-    This function can
;                                   handle the exception.
;
;    ExceptionContinueSearch-       This function cannot
;                                   handle the exception.

                    .code
sPtr                textequ <(scopeTable ptr [r8])>
genericSEHfltr      proc

; Assume failure:

                    mov     eax, ExceptionContinueSearch
                    
; See if the exception type matches:

                    cmp     edx, sPtr.exType
                    jne     cantHandle
                    
; See if we're within the range of the protected
; try block:

                    cmp     rcx, sPtr.beginAddress
                    jb      cantHandle
                    cmp     rcx, sPtr.endAddress
                    jae     cantHandle
            
; This handler can handle the exception:

                    mov     eax, ExceptionContinueExecution

cantHandle:         ret     
genericSEHfltr      endp    




; genericLSH is a generic language-specific 
; handler that gets called by a procedure-specific
; handler. This procedure has the following
; parameters:
;
;  RCX:     pExceptionRecord (Windows supplies this)
;  RDX:     pEstablisherFrame (Windows supplies this)
;  R8:      pDispatcherContext (Windows supplies this)
;  R9:      pDispatcherContext (Windows supplies this)
;  R10:     pScopeTable (procedure-specific handler
;                        supplies this).

rcxER       textequ <(ExceptionRecord_t ptr [rcx])>
rcxDC       textequ <(DispatcherContext ptr [rcx])>
r9DC        textequ <(DispatcherContext ptr [r9])>
raxDC       textequ <(DispatcherContext ptr [rax])>
r8SC        textequ <(scopeTable ptr [r8])>
raxRT       textequ <(RUNTIME_FUNCTION ptr [rax])>

genericLSH  proc    pExceptionRecord:  qword,  \
                    pEstablisherFrame: qword,  \
                    pContextRecord:    qword,  \
                    pDispatcherContext:qword,  \
                    pScopeTable:       qword
                    
; Note: 1 qword + 2 dwords leave stack 16-byte aligned.

            local   scPtr:             qword
            local   count[2]:          dword
            
; shadowStorage must always be last so it is on the
; bottom of the stack:

            local   shadowStorage[64]:byte
            
            
; Try and use volatile registers throughout this
; procedure so we don't have to deal with saving
; and restoring them.

            mov     pExceptionRecord, rcx
            mov     pEstablisherFrame, rdx
            mov     pContextRecord, r8
            mov     pDispatcherContext, r9
            mov     pScopeTable, r10
                    
            mov     rcx, pExceptionRecord
            cmp     rcxER.ExceptionFlags, EXCEPTION_NONCONTINUABLE
            jne     isCont
            
; If it's non-continuable (why are we even here?), 
; then quit the process

ExitPgm:    mov     rcx, 1
            call    ExitProcess

; Bail if this is an unwind operation; this 
; procedure-specific handler doesn't deal with that.

isCont:     mov     eax, ExceptionContinueSearch
            test    dword ptr rcxER.ExceptionFlags, EXCEPTION_UNWIND
            jnz     exitProcSpec

; Set up for the search loop. On entry to this procedure,
; pScopeTable contained a pointer to a ScopeTable
; record that has the following form:
;
;  count:dword
;
;  scArray scopeTable count dup ({})
;
; (scArray is a variable-sized array with 1 or more
;  scopeTable records in it).
;
; Each scopeTable entry as the following:
;
; scopeTable        struct
;   exType          dword    2 dup (?) ;Type is in 1st
;   beginAddress    qword    ?         ;  dword. 2nd dword
;   endAddress      qword    ?         ;  is padding.
;   filterAddress   qword    ?
;   jumpTarget      qword    ?
; scopeTable        ends
; 
            mov     rcx, pExceptionRecord
            mov     edx, rcxER.ExceptionCode
            mov     rcx, pDispatcherContext
            mov     rcx, rcxDC.ControlPc
            mov     r8, pScopeTable
            mov     eax, [r8]       ;Get element count
            add     r8, 4           ;Point at 1st element
            mov     scPtr, r8
            mov     count, eax

; For each element in the array of scopeTable elements,
; call the filter function (pointed at by the
; "filterAddress" field of that array element) to 
; see if the current exception is within the block
; of code specified by this scopeTable item.        
;
; Windows ABI parameters for this:
;
;  filterFunc( RIP, exception code, scopeTableEntry )
;              ECX       RDX             R8
;
; If the filter function returns failure (it doesn't
; handle the current exception), then repeat the
; loop until we run out of array elements or one
; of them can handle the exception.
;
; Note: this code works under the assumption that the
; filter function does not modify any registers!

searchScps: 
            call    r8SC.filterAddress
            cmp     rax, ExceptionContinueExecution
            je      HandleException
            add     r8, sizeof scopeTable
            dec     count
            jnz     searchScps
            
; Quit if we got ExceptionContinueSearch back and we
; ran out of entries in the scopeTable array:

            jmp     exitProcSpec  

; Call RtlUnwind to unwind the stack and transfer
; control to the actual exception handler code.
;
; RtlUnwind has the following parameters:
;
; RCX: target frame (pEstablisherFrame passed to this code)
; RDX: Address of the handler code (found at
;           r8SC.jumpTarget)
; R8:  pExceptionRecord (passed to this code)
; R9:  Return value that is passed on to the
;      exception handler in RAX.

HandleException:
            mov     rcx, pDispatcherContext
            mov     rdx, r8SC.jumpTarget
            mov     r8, pExceptionRecord
            mov     rcx, pEstablisherFrame
            
; This code is going to pass the pointer to
; the exception record to the handler, as that's
; probably the most useful piece of information
; it can have. Remember, this value shows
; up in RAX in the actual exception handler.

            mov     r9, r8
            call    RtlUnwind
            
; The code should never return from RtlUnwindEx
; unless there was an error. In that case,
; abort the program.

            mov     rcx, 1
            call    ExitProcess 
            
            
; Exit here if this procedure can't handle
; the exception.    
            
exitProcSpec:
            ret     
genericLSH  endp 
            end     