; Listing7-4.asm
;
; A program that demonstrates throwning
; an exception with the raiseException
; API call.


            option     casemap:none
            option     dotname              
            .nolist
            include    aoalib.inc
            include    aoaExcepts.inc
            includelib aoalib.lib
            .list
            
            .const

; Program title:

             align  word
ttlStr       byte   "Listing 7-4", 0

            .data
            
originalExceptContext Context {} 
            
LSHRtnVal   qword   0


            .code
            
; Once you use "option epilogue:xxxx" the default epilogue
; code seems to be broken. Just turn it off and generate
; epilogs manually:

            
; tstSEHfltr-
;
; This procedure must be a LEAF function!
; This function must not modify any register values other
; than EAX (the return value).
;
; Passed the following arguments:
;
;    RCX:   RIP value of interrupted instruction.
;    EDX:   Exception code.
;    R8:    Pointer to scopeTable record that
;           this function is to test.
;
; Return values (in RAX):
;
;    ExceptionContinueExecution-    This function can
;                                   handle the exception.
;
;    ExceptionContinueSearch-       This function cannot
;                                   handle the exception.

sPtr        textequ <(scopeTable ptr [r8])>
tstSEHfltr  proc

; Assume failure:

            mov     eax, ExceptionContinueSearch
            cmp     edx, sPtr.exType
            jne     cantHandle
            cmp     rcx, sPtr.beginAddress
            jb      cantHandle
            cmp     rcx, sPtr.endAddress
            jae     cantHandle
            
; This handler can handle the exception:

            mov     eax, ExceptionContinueExecution

cantHandle: ret     
tstSEHfltr  endp    


; genericLSH is a generic language-specific 
; handler that gets called by a procedure-specific
; handler. This procedure has the following
; parameters:
;
;  RCX:     pExceptionRecord (Windows supplies this)
;  RDX:     pEstablisherFrame (Windows supplies this)
;  R8:      pDispatcherContext (Windows supplies this)
;  R9:      pDispatcherContext (Windows supplies this)
;  R10:     pScopeTable (procedure-specific handler
;                        supplies this).

rcxER       textequ <(ExceptionRecord_t ptr [rcx])>
rcxDC       textequ <(DispatcherContext ptr [rcx])>
r8SC        textequ <(scopeTable ptr [r8])>
raxRT       textequ <(RUNTIME_FUNCTION ptr [rax])>

genericLSH  proc    pExceptionRecord:  qword,  \
                    pEstablisherFrame: qword,  \
                    pContextRecord:    qword,  \
                    pDispatcherContext:qword,  \
                    pScopeTable:       qword
                    
; Note: 1 qword + 2 dwords leave stack 16-byte aligned.

            local   scPtr:             qword
            local   count[2]:          dword
            
; shadowStorage must always be last so it is on the
; bottom of the stack:

            local   shadowStorage[64]:byte
            
            
; Try and use volatile registers throughout this
; procedure so we don't have to deal with saving
; and restoring them.

            mov     pExceptionRecord, rcx
            mov     pEstablisherFrame, rdx
            mov     pContextRecord, r8
            mov     pDispatcherContext, r9
            mov     pScopeTable, r10
                    
            mov     rcx, pExceptionRecord
            cmp     rcxER.ExceptionFlags, EXCEPTION_NONCONTINUABLE
            jne     isCont
            
; If it's non-continuable (why are we even here?), 
; then quit the process

ExitPgm:    mov     rcx, 1
            call    ExitProcess

; Bail if this is an unwind operation; this 
; procedure-specific handler doesn't deal with that.

isCont:     mov     eax, ExceptionContinueSearch
            test    dword ptr rcxER.ExceptionFlags, EXCEPTION_UNWIND
            jnz     exitProcSpec

; Set up for the search loop. On entry to this procedure,
; pScopeTable contained a pointer to a ScopeTable
; record that has the following form:
;
;  count:dword
;
;  scArray scopeTable count dup ({})
;
; (scArray is a variable-sized array with 1 or more
;  scopeTable records in it).
;
; Each scopeTable entry as the following:
;
; scopeTable        struct
;   exType          dword    2 dup (?) ;Type is in 1st
;   beginAddress    qword    ?         ;  dword. 2nd dword
;   endAddress      qword    ?         ;  is padding.
;   filterAddress   qword    ?
;   jumpTarget      qword    ?
; scopeTable        ends
; 
            mov     rcx, pExceptionRecord
            mov     edx, rcxER.ExceptionCode
            mov     rcx, pDispatcherContext
            mov     rcx, rcxDC.ControlPc
            mov     r8, pScopeTable
            mov     eax, [r8]       ;Get element count
            add     r8, 4           ;Point at 1st element
            mov     scPtr, r8
            mov     count, eax

; For each element in the array of scopeTable elements,
; call the filter function (pointed at by the
; "filterAddress" field of that array element) to 
; see if the current exception is within the block
; of code specified by this scopeTable item.        
;
; Windows ABI parameters for this:
;
;  filterFunc( RIP, exception code, scopeTableEntry )
;              ECX       RDX             R8
;
; If the filter function returns failure (it doesn't
; handle the current exception), then repeat the
; loop until we run out of array elements or one
; of them can handle the exception.
;
; Note: this code works under the assumption that the
; filter function does not modify any registers!

searchScps: 
            call    r8SC.filterAddress
            cmp     rax, ExceptionContinueExecution
            je      HandleException
            add     r8, sizeof scopeTable
            dec     count
            jnz     searchScps
            
; Quit if we got ExceptionContinueSearch back and we
; ran out of entries in the scopeTable array:

            jmp     exitProcSpec  

; Call RtlUnwind to unwind the stack and transfer
; control to the actual exception handler code.
;
; RtlUnwind has the following parameters:
;
; RCX: target frame (pEstablisherFrame passed to this code)
; RDX: Address of the handler code (found at
;           r8SC.jumpTarget)
; R8:  pExceptionRecord (passed to this code)
; R9:  Return value that is passed on to the
;      exception handler in RAX.

HandleException:
            mov     rcx, pDispatcherContext
            mov     rdx, r8SC.jumpTarget
            mov     r8, pExceptionRecord
            mov     rcx, pEstablisherFrame
            
; This code is going to pass the pointer to
; the exception record to the handler, as that's
; probably the most useful piece of information
; it can have. Remember, this value shows
; up in RAX in the actual exception handler.

            mov     r9, r8
            call    RtlUnwind
            
; The code should never return from RtlUnwindEx
; unless there was an error. In that case,
; abort the program.

            mov     rcx, 1
            call    ExitProcess 
            
            
; Exit here if this procedure can't handle
; the exception.    
            
exitProcSpec:
            ret     
genericLSH  endp 

            

; The following is a procedure-specific handler
; for the tstSEH procedure.

procSpec    proc
            lea     r10, tSEHST
            jmp     genericLSH
procSpec    endp


; tstSEH-
;
; This procedure tests exception handling code in
; assembly language.
            
tstSEH      xproc   (procSpec, <rax, rbx, rcx>)
                    
            local   shadowStorage[40]:byte
            
; Begin the protected block:
;
; try
     
startTry::
            int     3       ;Cause an exception
            
; This will never execute, but inlucde the code just
; to show the structure of the try..except..endtry
; code.
            call    finallyCode     ;Execute finally block
            jmp     endTryCode
endProtected:

; Here's the code that executes whenever a memory
; access violation occurs in the protected block.
;
; Note that the call to the finallyCode can happen
; before or after the body of the exception handler.
; Normally, it's best if it happens first (to
; guarantee execution). This may not be possible if
; executing that code could also lead to an exception.

; exception( ex(brkpt) )

brkPtExc::
            push    rax     ;Must preserve rax
            push    rax     ;16-byte align stack
            call    finallyCode
            pop     rax     ;Restore RAX
            pop     rax     ;16-byte align stack
            
; Upon entry into this code, RAX contains a pointer
; to the DispatcherContext record. Print a couple
; of useful values from that record:

            mov     edx, [rax].ExceptionRecord_t.ExceptionCode
            call    print
            byte    "Breakpoint encountered (0x%x)", nl, 0
            jmp     endTryCode
            
; Here's the finally code; This corresponds to the
; 

finallyCode:
            call    print
            byte    "Executed finally code", nl, 0
            ret     0
            

; Write an appropriate scopeTable record to the
; .xdata segment.
                            
.xdata      segment     read align(8)
tSEHST      dword       1
            scopeTable  {{ex(brkpt),0}, startTry, endProtected, tstSEHfltr, brkPtExc}
.xdata      ends

; This is where the code exists the protected loop.
; 
; endtry
           
endTryCode:
tstSEH      endxp()


        
;--------------------------------------------------
;
; Here is the main assembly language function.

            
            public  asmMain
            
            align       8
            qword       ?   ;Also for hot patching

; For the main assembly procedure, preserve all the 
; non-volatile registers. It would be cool to save  
; the XMM non-volatile registers, too, but there isn't 
; room in the prolog for them.
; 
; Had this code used a biased frame pointer (RBP) value, 
; it could have gotten away with saving them. Epilog 
; and prolog  macros don't allow that, however. So this 
; code preserves what it can (the general-purpose 64-bit 
; non-volatile registers).
            
;asmMain     proc        <rbx, rsi, rdi, r12, r13, r14, r15>        \
;                       frame,                              \
;                       parm:qword


asmMain     xproc   (frame, <rbx, rsi, rdi, r12, r13, r14, r15>)
            local       shadowStorage[64]:byte
            

; Call the procedure the demonstrates exception handling:

            call    print
            byte    "Calling tstSEH", nl, 0
            
; Call the procedure that raises an exception:

            call    tstSEH
            
            call    print
            byte    "Back from tstSEH", nl, 0
            
asmMain     endxp()
            end
